from contextlib import contextmanager
from typing import Optional, List

from .base import BaseSubAPI
from .. import ser
from ..sess import eval_lua_method_factory, lua_context_object


class SeekMixin:
    def seek(self, whence: str = None, offset: int = None) -> int:
        # whence: set, cur, end
        rp = self._method('seek', ser.nil_encode(whence), offset)
        rp.check_nil_error()
        return rp.take_int()


class ReadMixin:
    def _take(self, rp):
        raise NotImplementedError

    def read(self, count: int = 1) -> Optional[str]:
        return self._take(self._method('read', count))

    def readLine(self, withTrailing: bool = False) -> Optional[str]:
        return self._take(self._method('readLine', withTrailing))

    def readAll(self) -> Optional[str]:
        return self._take(self._method('readAll'))

    def __iter__(self):
        return self

    def __next__(self):
        line = self.readLine()
        if line is None:
            raise StopIteration
        return line


class WriteMixin:
    def _put(self, t):
        raise NotImplementedError

    def write(self, text: str):
        return self._method('write', self._put(text)).take_none()

    def flush(self):
        return self._method('flush').take_none()


class ReadHandle(ReadMixin, BaseSubAPI):
    def _take(self, rp):
        return rp.take_option_unicode()


class BinaryReadHandle(ReadMixin, SeekMixin, BaseSubAPI):
    def _take(self, rp):
        return rp.take_option_bytes()


class WriteHandle(WriteMixin, BaseSubAPI):
    def _put(self, t: str) -> bytes:
        return t.encode('utf-8')

    def writeLine(self, text: str):
        return self.write(text + '\n')


class BinaryWriteHandle(WriteMixin, SeekMixin, BaseSubAPI):
    def _put(self, b: bytes) -> bytes:
        return b


method = eval_lua_method_factory('fs.')


__all__ = (
    'list',
    'exists',
    'isDir',
    'isReadOnly',
    'getDrive',
    'getSize',
    'getFreeSpace',
    'getCapacity',
    'makeDir',
    'move',
    'copy',
    'delete',
    'combine',
    'open',
    'find',
    'getDir',
    'getName',
    'isDriveRoot',
    'complete',
    'attributes',
)


def list(path: str) -> List[str]:
    return method('list', ser.encode(path)).take_list_of_strings()


def exists(path: str) -> bool:
    return method('exists', ser.encode(path)).take_bool()


def isDir(path: str) -> bool:
    return method('isDir', ser.encode(path)).take_bool()


def isReadOnly(path: str) -> bool:
    return method('isReadOnly', ser.encode(path)).take_bool()


def getDrive(path: str) -> Optional[str]:
    return method('getDrive', ser.encode(path)).take_option_string()


def getSize(path: str) -> int:
    return method('getSize', ser.encode(path)).take_int()


def getFreeSpace(path: str) -> int:
    return method('getFreeSpace', ser.encode(path)).take_int()


def getCapacity(path: str) -> int:
    return method('getCapacity', ser.encode(path)).take_int()


def makeDir(path: str):
    return method('makeDir', ser.encode(path)).take_none()


def move(fromPath: str, toPath: str):
    return method('move', ser.encode(fromPath), ser.encode(toPath)).take_none()


def copy(fromPath: str, toPath: str):
    return method('copy', ser.encode(fromPath), ser.encode(toPath)).take_none()


def delete(path: str):
    return method('delete', ser.encode(path)).take_none()


def combine(basePath: str, localPath: str) -> str:
    return method('combine', ser.encode(basePath), ser.encode(localPath)).take_string()


@contextmanager
def open(path: str, mode: str):
    '''
    Usage:

    with fs.open('filename', 'w') as f:
        f.writeLine('textline')

    with fs.open('filename', 'r') as f:
        for line in f:
            ...
    '''
    with lua_context_object(
        'fs.open(...)',
        (ser.encode(path), ser.encode(mode.replace('b', '') + 'b')),
        '{e}.close()',
    ) as var:
        if 'b' in mode:
            hcls = BinaryReadHandle if 'r' in mode else BinaryWriteHandle
        else:
            hcls = ReadHandle if 'r' in mode else WriteHandle
        yield hcls(var)


def find(wildcard: str) -> List[str]:
    return method('find', ser.encode(wildcard)).take_list_of_strings()


def getDir(path: str) -> str:
    return method('getDir', ser.encode(path)).take_string()


def getName(path: str) -> str:
    return method('getName', ser.encode(path)).take_string()


def isDriveRoot(path: str) -> bool:
    return method('isDriveRoot', ser.encode(path)).take_bool()


def complete(
    partialName: str, path: str, includeFiles: bool = None, includeDirs: bool = None,
) -> List[str]:
    return method(
        'complete', ser.encode(partialName), ser.encode(path), includeFiles, includeDirs,
    ).take_list_of_strings()


def attributes(path: str) -> dict:
    tp = method('attributes', ser.encode(path)).take_dict((
        b'created',
        b'modification',
        b'isDir',
        b'size',
    ))
    r = {}
    r['created'] = tp.take_int()
    r['modification'] = tp.take_int()
    r['isDir'] = tp.take_bool()
    r['size'] = tp.take_int()
    return r



from types import ModuleType

from ..errors import LuaException
from ..lua import lua_string
from ..sess import eval_lua


__all__ = (
    'import_file',
    'is_commands',
    'is_multishell',
    'is_turtle',
    'is_pocket',
    'eval_lua',
    'LuaException',
)


def import_file(path: str, relative_to: str = None):
    mod = ModuleType(path)
    mod.__file__ = path
    path_expr = lua_string(path)
    if relative_to is not None:
        path_expr = 'fs.combine(fs.getDir({}), {})'.format(
            lua_string(relative_to),
            path_expr,
        )
    source = eval_lua('''
local p = {}
if not fs.exists(p) then return nil end
if fs.isDir(p) then return nil end
f = fs.open(p, "r")
local src = f.readAll()
f.close()
return src
'''.lstrip().format(
        path_expr,
    )).take_option_string()
    if source is None:
        raise ImportError('File not found: {}'.format(path))
    cc = compile(source, mod.__name__, 'exec')
    exec(cc, vars(mod))
    return mod


def is_commands() -> bool:
    return eval_lua('return commands ~= nil').take_bool()


def is_multishell() -> bool:
    return eval_lua('return multishell ~= nil').take_bool()


def is_turtle() -> bool:
    return eval_lua('return turtle ~= nil').take_bool()


def is_pocket() -> bool:
    return eval_lua('return pocket ~= nil').take_bool()



from typing import Optional

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('multishell.')


__all__ = (
    'getCurrent',
    'getCount',
    'launch',
    'setTitle',
    'getTitle',
    'setFocus',
    'getFocus',
)


def getCurrent() -> int:
    return method('getCurrent').take_int()


def getCount() -> int:
    return method('getCount').take_int()


def launch(environment: dict, programPath: str, *args: str) -> int:
    args = tuple(ser.encode(a) for a in args)
    return method('launch', environment, ser.encode(programPath), *args).take_int()


def setTitle(tabID: int, title: str):
    return method('setTitle', tabID, ser.encode(title)).take_none()


def getTitle(tabID: int) -> Optional[str]:
    return method('getTitle', tabID).take_option_string()


def setFocus(tabID: int) -> bool:
    return method('setFocus', tabID).take_bool()


def getFocus() -> int:
    return method('getFocus').take_int()



from dataclasses import dataclass
from typing import Dict, Optional, List, Tuple, Type, Any, Union
import json, ast
from .mixins import TermMixin, TermTarget
from .. import ser
from ..lua import LuaNum, lua_string
from ..sess import eval_lua, eval_lua_method_factory
import json


def debytify(dict):
    newDict = ast.literal_eval(str(dict).replace("b'", "'").replace("'",'"'))
    return (newDict)


class BasePeripheral:
    # NOTE: is not LuaExpr, you can't pass peripheral as parameter
    # TODO: to fix this we can supply separate lua expr, result of .wrap()

    def __init__(self, lua_method_expr, *prepend_params):
        self._lua_method_expr = lua_method_expr
        self._prepend_params = prepend_params

    def _method(self, name, *params):
        code = 'return ' + self._lua_method_expr + '(...)'
        return eval_lua(code, *self._prepend_params, ser.encode(name), *params)


class CCDrive(BasePeripheral):
    def isDiskPresent(self) -> bool:
        return self._method('isDiskPresent').take_bool()

    def getDiskLabel(self) -> Optional[str]:
        return self._method('getDiskLabel').take_option_string()

    def setDiskLabel(self, label: Optional[str]):
        return self._method('setDiskLabel', ser.nil_encode(label)).take_none()

    def hasData(self) -> bool:
        return self._method('hasData').take_bool()

    def getMountPath(self) -> Optional[str]:
        return self._method('getMountPath').take_option_string()

    def hasAudio(self) -> bool:
        return self._method('hasAudio').take_bool()

    def getAudioTitle(self) -> Optional[Union[bool, str]]:
        return self._method('getAudioTitle').take_option_string_bool()

    def playAudio(self):
        return self._method('playAudio').take_none()

    def stopAudio(self):
        return self._method('stopAudio').take_none()

    def ejectDisk(self):
        return self._method('ejectDisk').take_none()

    def getDiskID(self) -> Optional[int]:
        return self._method('getDiskID').take_option_int()


class CCMonitor(BasePeripheral, TermMixin):
    def getTextScale(self) -> int:
        return self._method('getTextScale').take_int()

    def setTextScale(self, scale: int):
        return self._method('setTextScale', scale).take_none()


class ComputerMixin:
    def turnOn(self):
        return self._method('turnOn').take_none()

    def shutdown(self):
        return self._method('shutdown').take_none()

    def reboot(self):
        return self._method('reboot').take_none()

    def getID(self) -> int:
        return self._method('getID').take_int()

    def getLabel(self) -> Optional[str]:
        return self._method('getLabel').take_option_string()

    def isOn(self) -> bool:
        return self._method('isOn').take_bool()


class CCComputer(BasePeripheral, ComputerMixin):
    pass


class CCTurtle(BasePeripheral, ComputerMixin):
    pass


@dataclass
class ModemMessage:
    reply_channel: int
    content: Any
    distance: LuaNum


class ModemMixin:
    def isOpen(self, channel: int) -> bool:
        return self._method('isOpen', channel).take_bool()

    def open(self, channel: int):
        return self._method('open', channel).take_none()

    def close(self, channel: int):
        return self._method('close', channel).take_none()

    def closeAll(self):
        return self._method('closeAll').take_none()

    def transmit(self, channel: int, replyChannel: int, message: Any):
        return self._method('transmit', channel, replyChannel, message).take_none()

    def isWireless(self) -> bool:
        return self._method('isWireless').take_bool()

    @property
    def _side(self):
        return self._prepend_params[0]

    def receive(self, channel: int):
        from .os import captureEvent

        if self.isOpen(channel):
            raise Exception('Channel is busy')

        self.open(channel)
        try:
            for evt in captureEvent('modem_message'):
                if evt[0] != self._side:
                    continue
                if evt[1] != channel:
                    continue
                yield ModemMessage(*evt[2:])
        finally:
            self.close(channel)


class CCWirelessModem(BasePeripheral, ModemMixin):
    pass


class CCWiredModem(BasePeripheral, ModemMixin):
    def getNameLocal(self) -> Optional[str]:
        return self._method('getNameLocal').take_option_string()

    def getNamesRemote(self) -> List[str]:
        return self._method('getNamesRemote').take_list_of_strings()

    def getTypeRemote(self, peripheralName: str) -> Optional[str]:
        return self._method('getTypeRemote', ser.encode(peripheralName)).take_option_string()

    def isPresentRemote(self, peripheralName: str) -> bool:
        return self._method('isPresentRemote', ser.encode(peripheralName)).take_bool()

    def wrapRemote(self, peripheralName: str) -> Optional[BasePeripheral]:
        # use instead getMethodsRemote and callRemote
        # NOTE: you can also use peripheral.wrap(peripheralName)

        ptype = self.getTypeRemote(peripheralName)
        if ptype is None:
            return None

        return TYPE_MAP[ptype](
            self._lua_method_expr, *self._prepend_params,
            b'callRemote', ser.encode(peripheralName),
        )

    # NOTE: for TermTarget use peripheral.get_term_target(peripheralName)


class CCPrinter(BasePeripheral):
    def newPage(self) -> bool:
        return self._method('newPage').take_bool()

    def endPage(self) -> bool:
        return self._method('endPage').take_bool()

    def write(self, text: str):
        return self._method('write', ser.dirty_encode(text)).take_none()

    def setCursorPos(self, x: int, y: int):
        return self._method('setCursorPos', x, y).take_none()

    def getCursorPos(self) -> Tuple[int, int]:
        rp = self._method('getCursorPos')
        return tuple(rp.take_int() for _ in range(2))

    def getPageSize(self) -> Tuple[int, int]:
        rp = self._method('getPageSize')
        return tuple(rp.take_int() for _ in range(2))

    def setPageTitle(self, title: str):
        return self._method('setPageTitle', ser.encode(title)).take_none()

    def getPaperLevel(self) -> int:
        return self._method('getPaperLevel').take_int()

    def getInkLevel(self) -> int:
        return self._method('getInkLevel').take_int()


class CCSpeaker(BasePeripheral):
    def playNote(self, instrument: str, volume: int = 1, pitch: int = 1) -> bool:
        # instrument:
        # https://minecraft.gamepedia.com/Note_Block#Instruments
        # bass
        # basedrum
        # bell
        # chime
        # flute
        # guitar
        # hat
        # snare
        # xylophone
        # iron_xylophone
        # pling
        # banjo
        # bit
        # didgeridoo
        # cow_bell

        # volume 0..3
        # pitch 0..24
        return self._method('playNote', ser.encode(instrument), volume, pitch).take_bool()

    def playSound(self, sound: str, volume: int = 1, pitch: int = 1) -> bool:
        # volume 0..3
        # pitch 0..2
        return self._method('playSound', ser.encode(sound), volume, pitch).take_bool()


class CCCommandBlock(BasePeripheral):
    def getCommand(self) -> str:
        return self._method('getCommand').take_string()

    def setCommand(self, command: str):
        return self._method('setCommand', ser.encode(command)).take_none()

    def runCommand(self):
        return self._method('runCommand').check_bool_error()


class CCWorkbench(BasePeripheral):
    def craft(self, quantity: int = 64):
        return self._method('craft', quantity).check_bool_error()


class CCInventory(BasePeripheral):
    def getItemDetail(self, slot: int) -> Optional[dict]:
        return self._method('getItemDetail', slot).take()

    def list(self) -> Dict[int, dict]:
        return debytify(self._method('list').take_dict())

    def pullItems(self, fromName: str, fromSlot: int, limit: int = None, toSlot: int = None) -> int:
        return self._method('pullItems', ser.encode(fromName), fromSlot, limit, toSlot).take_int()

    def pushItems(self, toName: str, fromSlot: int, limit: int = None, toSlot: int = None) -> int:
        return self._method('pushItems', ser.encode(toName), fromSlot, limit, toSlot).take_int()

    def size(self) -> int:
        return self._method('size').take_int()


class CCBigReactors(BasePeripheral):
    def getEnergyStats(self):
        d = self._method('getEnergyStats').take_dict()
        return debytify(d)

    def setActive(self, state: bool):
        return self._method('setActive', state).take_none()

    def getActive(self):
        return self._method('getActive').take_bool()


class CCPowah(BasePeripheral):
    def getEnergyStats(self):
        energy = self._method('getEnergy').take_int()
        capacity = self._method('getEnergyCapacity').take_int()
        return {"energyStored": energy, "capacity": capacity}


class AnyPeripheral(BasePeripheral):
    def getReturn(self, newmethod, datatype):
        if datatype == "int":
            return self._method(newmethod).take_int()
        if datatype == "bool":
            return self._method(newmethod).take_bool()
        if datatype == "dict":
            return debytify(self._method(newmethod).take_dict())
        return self._method(newmethod).take_none()

    def set(self, newmethod: str, state: bool):
        return self._method(newmethod, state).take_none()

TYPE_MAP = {}

method = eval_lua_method_factory('peripheral.')

__all__ = (
    'BasePeripheral',  # exposed for subclassing & registerType
    'CCInventory',  # exposed for registerType for 3rdparty mod inventory-like entities
    'isPresent',
    'getType',
    'getNames',
    'wrap',
    'registerType',
    'get_term_target',
    'registerPeripheral',
)


def isPresent(side: str) -> bool:
    return method('isPresent', ser.encode(side)).take_bool()


def getType(side: str) -> Optional[str]:
    return method('getType', ser.encode(side)).take_option_string()


def getNames() -> List[str]:
    return method('getNames').take_list_of_strings()


# use instead getMethods and call
def wrap(side: str) -> Optional[BasePeripheral]:
    ptype = getType(side)
    if ptype is None:
        return None

    m = 'peripheral.call'

    side = ser.encode(side)
    if ptype == 'modem':
        if method('call', side, b'isWireless').take_bool():
            return CCWirelessModem(m, side)
        else:
            return CCWiredModem(m, side)
    else:
        return TYPE_MAP[ptype](m, side)


def registerType(ptype: str, pcls: Type[BasePeripheral]):
    TYPE_MAP[ptype] = pcls


def get_term_target(side: str) -> TermTarget:
    return TermTarget('peripheral.wrap({})'.format(
        lua_string(side),
    ))


def registerPeripheral(peipheralcode):
    registerType(peipheralcode, AnyPeripheral)


registerType('drive', CCDrive)
registerType('monitor', CCMonitor)
registerType('computer', CCComputer)
registerType('turtle', CCTurtle)
registerType('printer', CCPrinter)
registerType('speaker', CCSpeaker)
registerType('command', CCCommandBlock)
registerType('workbench', CCWorkbench)
registerType('BigReactors-Reactor', CCBigReactors)
registerType('powah:energy_cell', CCPowah)
for k in [
    'chest',
    'furnace',
    'barrel',
    'hopper',
    'dropper',
    'dispenser',
    'blast_furnace',
    'smoker',
    'shulker_box',
    'brewing_stand',
]:
    registerType('minecraft:' + k, CCInventory)



from typing import Optional

from .. import ser
from ..sess import eval_lua, eval_lua_method_factory


method = eval_lua_method_factory('keys.')


__all__ = (
    'getCode',
    'getName',
)


def getCode(name: str) -> Optional[int]:
    # replaces properties
    # keys.space → keys.getCode('space')
    return eval_lua('''
local k = ...
if type(keys[k]) == 'number' then
    return keys[k]
end
return nil''', ser.encode(name)).take_option_int()


def getName(code: int) -> Optional[str]:
    return method('getName', code).take_option_string()



from typing import List, Union

from .. import ser
from ..lua import LuaNum
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('textutils.')


__all__ = (
    'slowWrite',
    'slowPrint',
    'formatTime',
    'tabulate',
    'pagedTabulate',
    'pagedPrint',
    'complete',
)


def slowWrite(text: str, rate: LuaNum = None):
    return method('slowWrite', ser.dirty_encode(text), rate).take_none()


def slowPrint(text: str, rate: LuaNum = None):
    return method('slowPrint', ser.dirty_encode(text), rate).take_none()


def formatTime(time: LuaNum, twentyFourHour: bool = None) -> str:
    return method('formatTime', time, twentyFourHour).take_string()


def _prepareTab(rows_and_colors):
    r = []
    for item in rows_and_colors:
        if isinstance(item, int):
            r.append(item)
        else:
            r.append([ser.dirty_encode(x) for x in item])
    return r


def tabulate(*rows_and_colors: Union[List[str], int]):
    return method('tabulate', *_prepareTab(rows_and_colors)).take_none()


def pagedTabulate(*rows_and_colors: Union[List[str], int]):
    return method('pagedTabulate', *_prepareTab(rows_and_colors)).take_none()


def pagedPrint(text: str, freeLines: int = None) -> int:
    return method('pagedPrint', ser.dirty_encode(text), freeLines).take_int()


def complete(partial: str, possible: List[str]) -> List[str]:
    return [p[len(partial):] for p in possible if p.startswith(partial)]


# Questionable to implement
# serialize
# unserialize

# Will not implement, use pythonic equivalents
# serializeJSON
# unserializeJSON
# urlEncode
# json_null
# empty_json_array



from typing import Tuple, List, Optional

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('commands.')


__all__ = (
    'exec',
    'list',
    'getBlockPosition',
    'getBlockInfo',
    'getBlockInfos',
)


def exec(command: str) -> Tuple[bool, List[str], Optional[int]]:
    rp = method('exec', ser.encode(command))
    success = rp.take_bool()
    log = rp.take_list_of_strings()
    n = rp.take_option_int()
    return success, log, n


def list() -> List[str]:
    return method('list').take_list_of_strings()


def getBlockPosition() -> Tuple[int, int, int]:
    rp = method('getBlockPosition')
    return tuple(rp.take_int() for _ in range(3))


def getBlockInfo(x: int, y: int, z: int) -> dict:
    return method('getBlockInfo', x, y, z).take_dict()


def getBlockInfos(x1: int, y1: int, z1: int, x2: int, y2: int, z2: int) -> List[dict]:
    return method('getBlockInfos', x1, y1, z1, x2, y2, z2).take_list()



from typing import Union

from . import ser


LuaTable = Union[list, dict]
LuaNum = Union[int, float]


class LuaExpr:
    def get_expr_code(self):
        raise NotImplementedError


_tmap = {
    '\\': '\\\\',
    '\a': '\\a',
    '\b': '\\b',
    '\f': '\\f',
    '\n': '\\n',
    '\r': '\\r',
    '\t': '\\t',
    '\v': '\\v',
    '"': '\\"',
    "'": "\\'",
    '[': '\\[',
    ']': '\\]',
}
_tmap = {ord(c): r for c, r in _tmap.items()}


def lua_string(v):
    if isinstance(v, bytes):
        v = ser.decode(v)
    return '"{}"'.format(v.translate(_tmap))



from ..sess import CCGreenlet, get_current_greenlet


__all__ = (
    'waitForAny',
    'waitForAll',
)


def waitForAny(*task_fns):
    pgl = get_current_greenlet().cc_greenlet
    sess = pgl._sess

    gs = [CCGreenlet(fn) for fn in task_fns]
    for g in gs:
        g.defer_switch()

    try:
        sess._server_greenlet.switch()
    finally:
        pgl.detach_children()


def waitForAll(*task_fns):
    pgl = get_current_greenlet().cc_greenlet
    sess = pgl._sess

    gs = [CCGreenlet(fn) for fn in task_fns]
    for g in gs:
        g.defer_switch()

    try:
        for _ in range(len(task_fns)):
            sess._server_greenlet.switch()
    finally:
        pgl.detach_children()



from contextlib import contextmanager
from typing import Tuple

from ..sess import eval_lua_method_factory, lua_context_object
from .base import BaseSubAPI
from .mixins import TermMixin, TermTarget


class CCWindow(BaseSubAPI, TermMixin):
    def setVisible(self, visibility: bool):
        return self._method('setVisible', visibility).take_none()

    def redraw(self):
        return self._method('redraw').take_none()

    def restoreCursor(self):
        return self._method('restoreCursor').take_none()

    def getPosition(self) -> Tuple[int, int]:
        rp = self._method('getPosition')
        return tuple(rp.take_int() for _ in range(2))

    def reposition(self, x: int, y: int, width: int = None, height: int = None, parent: TermTarget = None):
        return self._method('reposition', x, y, width, height, parent).take_none()

    def getLine(self, y: int) -> Tuple[str, bytes, bytes]:
        rp = self._method('getLine', y)
        return rp.take_string(), rp.take_bytes(), rp.take_bytes()

    def get_term_target(self) -> TermTarget:
        return TermTarget(self.get_expr_code())


method = eval_lua_method_factory('window.')


__all__ = (
    'create',
)


@contextmanager
def create(
    parentTerm: TermTarget, x: int, y: int, width: int, height: int, visible: bool = None,
) -> CCWindow:
    with lua_context_object(
        'window.create(...)',
        (parentTerm, x, y, width, height, visible),
    ) as var:
        yield CCWindow(var)



from typing import Optional, Union

from .. import ser
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('disk.')


__all__ = (
    'isPresent',
    'hasData',
    'getMountPath',
    'setLabel',
    'getLabel',
    'getID',
    'hasAudio',
    'getAudioTitle',
    'playAudio',
    'stopAudio',
    'eject',
)


def isPresent(side: str) -> bool:
    return method('isPresent', ser.encode(side)).take_bool()


def hasData(side: str) -> bool:
    return method('hasData', ser.encode(side)).take_bool()


def getMountPath(side: str) -> Optional[str]:
    return method('getMountPath', ser.encode(side)).take_option_string()


def setLabel(side: str, label: Optional[str]):
    return method('setLabel', ser.encode(side), ser.nil_encode(label)).take_none()


def getLabel(side: str) -> Optional[str]:
    return method('getLabel', ser.encode(side)).take_option_string()


def getID(side: str) -> Optional[int]:
    return method('getID', ser.encode(side)).take_option_int()


def hasAudio(side: str) -> bool:
    return method('hasAudio', ser.encode(side)).take_bool()


def getAudioTitle(side: str) -> Optional[Union[bool, str]]:
    return method('getAudioTitle', ser.encode(side)).take_option_string_bool()


def playAudio(side: str):
    return method('playAudio', ser.encode(side)).take_none()


def stopAudio(side: str):
    return method('stopAudio', ser.encode(side)).take_none()


def eject(side: str):
    return method('eject', ser.encode(side)).take_none()



from typing import Tuple, Optional

from ..lua import LuaNum
from ..sess import eval_lua_method_factory


method = eval_lua_method_factory('gps.')


__all__ = (
    'CHANNEL_GPS',
    'locate',
)


CHANNEL_GPS = 65534


def locate(timeout: LuaNum = None, debug: bool = None) -> Optional[Tuple[LuaNum, LuaNum, LuaNum]]:
    rp = method('locate', timeout, debug)
    if rp.peek() is None:
        return None
    return tuple(rp.take_number() for _ in range(3))
